import copy
import json
from datetime import datetime
from unittest import mock

from django.test import TransactionTestCase
from django.urls import reverse
from freezegun import freeze_time

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager
from django_forest.utils.date import get_timezone

@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
@mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
class StatsStatsWithParametersViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json', ]

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = f"{reverse('django_forest:stats:statsWithParameters', kwargs={'resource': 'tests_question'})}?timezone=Europe%2FParis"
        self.empty_url = f"{reverse('django_forest:stats:statsWithParameters', kwargs={'resource': 'tests_waiter'})}?timezone=Europe%2FParis"
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless_here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    def test_get_value_count(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 3
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_value_sum(self, *args, **kwargs):
        body = {
            'aggregate': 'Sum',
            'aggregate_field': 'id',
            'collection': 'tests_question',
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 6
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_value_sum_empty(self, *args, **kwargs):
        body = {
            'aggregate': 'Sum',
            'aggregate_field': 'id',
            'collection': 'tests_waiter',
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.empty_url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 0
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2021, 6, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_yesterday(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"pub_date\",\"operator\":\"yesterday\",\"value\":null}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 1,
                'countPrevious': 2
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2021, 6, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_yesterday_aggregator(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"aggregator\":\"and\",\"conditions\":[{\"field\":\"pub_date\",\"operator\":\"yesterday\",\"value\":null},{\"field\":\"pub_date\",\"operator\":\"present\",\"value\":null}]}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 1,
                'countPrevious': 2
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2021, 7, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_month(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"pub_date\",\"operator\":\"previous_month\",\"value\":null}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 3,
                'countPrevious': 0
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2021, 7, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_quarter(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"pub_date\",\"operator\":\"previous_quarter\",\"value\":null}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 3,
                'countPrevious': 0
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2021, 7, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_quarter_to_date(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"pub_date\",\"operator\":\"previous_quarter_to_date\",\"value\":null}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 0,
                'countPrevious': 3
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    @freeze_time(
        lambda: datetime(2022, 7, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_value_count_previous_year(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"pub_date\",\"operator\":\"previous_year\",\"value\":null}",
            'query': None,
            'time_range': None,
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 3,
                'countPrevious': 0
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_objective_count(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'query': None,
            'time_range': None,
            'type': 'Objective'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'value': 3
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_pie_count(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'query': None,
            'group_by_field': 'pub_date',
            'type': 'Pie'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': '02/06/2021',
                    'value': 2
                },
                {
                    'key': '03/06/2021',
                    'value': 1
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_pie_count_group_by_related_field(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'query': None,
            'group_by_field': 'topic:name',
            'type': 'Pie'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': None,
                    'value': 3
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_pie_sum(self, *args, **kwargs):
        body = {
            'aggregate': 'Sum',
            'aggregate_field': 'id',
            'collection': 'tests_question',
            'query': None,
            'group_by_field': 'pub_date',
            'type': 'Pie'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': '02/06/2021',
                    'value': 3
                },
                {
                    'key': '03/06/2021',
                    'value': 3
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_count_day(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'time_range': 'Day',
            'group_by_date_field': 'pub_date',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': '02/06/2021',
                    'values': {
                        'value': 2
                    }
                },
                {
                    'label': '03/06/2021',
                    'values': {
                        'value': 1
                    }
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_count_week(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'time_range': 'Week',
            'group_by_date_field': 'pub_date',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': 'W22-2021',
                    'values': {
                        'value': 3
                    }
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_count_month(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'time_range': 'Month',
            'group_by_date_field': 'pub_date',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': 'Jun 21',
                    'values': {
                        'value': 3
                    }
                },
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_count_year(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'time_range': 'Year',
            'group_by_date_field': 'pub_date',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': '2021',
                    'values': {
                        'value': 3
                    }
                },
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_count_empty(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'filters': "{\"field\":\"id\",\"operator\":\"equal\",\"value\":0}",
            'time_range': 'Day',
            'group_by_date_field': 'pub_date',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': []
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_leaderboard_count(self, *args, **kwargs):
        body = {
            'aggregate': 'Count',
            'collection': 'tests_question',
            'label_field': 'question_text',
            'limit': 5,
            'query': None,
            'time_range': None,
            'relationship_field': 'choice_set',
            'type': 'Leaderboard'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': 'what is your favorite color?',
                    'value': 2
                },
                {
                    'key': 'do you like chocolate?',
                    'value': 1
                },
                {
                    'key': 'who is your favorite singer?',
                    'value': 0
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')



import json

from django_forest.utils.date import get_timezone
from .utils import ConditionsMixin


class FiltersMixin(ConditionsMixin):
    def get_filters(self, params, Model):
        filters = json.loads(params['filters'])

        tz = None
        if 'timezone' in params:
            tz = get_timezone(params['timezone'])

        if 'aggregator' in filters:
            return self.handle_aggregator(filters, Model, tz)

        return self.get_expression(filters, Model, tz)



from django_forest.resources.associations.views.count import CountView
from django_forest.resources.associations.views.csv import CsvView
from django_forest.resources.associations.views.list import ListView

__all__ = ['ListView', 'CountView', 'CsvView']



from django_forest.tests.models import Place
from django_forest.utils.collection import Collection

from django_forest.resources.utils.resource import ResourceView
from django_forest.resources.views import ListView, DetailView

class PlaceForest(Collection):
    def load(self):
        self.fields = [
            {
                'field': 'restaurant',
                "reference": "tests_restaurant.id",
                'type': 'String',
                "filter": self.filter_restaurant,
                "is_filterable": True,

            }
        ]

    def filter_restaurant(self, obj):
        if obj.restaurant:
            return obj.restaurant
        return None

Collection.register(PlaceForest, Place)



from django.http import JsonResponse
from django_forest.utils.forest_setting import get_forest_setting


class DeactivateCountMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def build_association_resource(self, resource, association):
        return f"{resource}:{association}"

    def is_deactivated(self, request, view_func, *args, **kwargs):
        is_count_request = request.resolver_match.url_name == 'count'
        deactivated_count = get_forest_setting(
            'DEACTIVATED_COUNT',
            list
        )
        resolver_kwargs = request.resolver_match.kwargs
        association_resource = resolver_kwargs.get('association_resource')
        resource = resolver_kwargs.get('resource')
        all_associations = resource
        if association_resource:
            all_associations = self.build_association_resource(resource, '*')
            resource = self.build_association_resource(resource, association_resource)

        return is_count_request and any([resource in deactivated_count, all_associations in deactivated_count])

    def process_view(self, request, view_func, *args, **kwargs):
        if self.is_deactivated(request, view_func, *args, **kwargs):
            return JsonResponse({"meta": {"count": "deactivated"}})
        return None



import re

import marshmallow as ma
from marshmallow.schema import SchemaMeta
from marshmallow_jsonapi import Schema, fields
from django.core.exceptions import FieldDoesNotExist
from django_forest.utils.models import Models
from django_forest.utils.type_mapping import get_type


TYPE_CHOICES = {
    'String': fields.Str,
    'Integer': fields.Integer,
    'Float': fields.Float,
    'Number': fields.Number,
    'Boolean': fields.Boolean,
    'Dateonly': fields.Date,
    'Date': fields.DateTime,
    'Time': fields.Time,
    'Json': fields.Raw,
}


class JsonApiSchema(type):
    _registry = {}

    def __new__(mcs, model_name, bases, attrs):
        klass = super(JsonApiSchema, mcs).__new__(mcs, model_name, bases, attrs)
        mcs._registry[model_name] = klass
        return klass

    @classmethod
    def get(mcs, model_name):
        model_name = f'{model_name}Schema'
        if model_name in mcs._registry:
            return mcs._registry[model_name]
        raise Exception(f'The {model_name} does not exist in the JsonApiSchema. Make sure you correctly set it.')


# Notice: handle metaclass conflict
class MarshmallowType(JsonApiSchema, SchemaMeta):
    pass


def get_type_name(name):
    return re.sub(r'(\w)([A-Z])', r'\1 \2', name)


def get_marshmallow_field(field, Model):
    # Notice, handle Model is None (Smart Collection)
    model_fields = [] if Model is None else Model._meta.get_fields()

    # Notice, handle smart field, default to type
    _type = next((get_type(x) for x in model_fields if x.name == field['field']), field['type'])

    if isinstance(_type, list):
        return fields.List(TYPE_CHOICES.get(_type[0], fields.Str)())
    return TYPE_CHOICES.get(_type, fields.Str)()


class DjangoRelationship(fields.Relationship):

    def _serialize(self, value, attr, obj):
        if value and self.many:
            value = value.all()
        return super(DjangoRelationship, self)._serialize(value, attr, obj)


def populate_attrs(collection, collection_name):
    Model = Models.get(collection_name)
    attrs = {}
    for field in collection['fields']:
        field_name = field['field']
        if field['reference'] is not None:
            related_name = field['reference'].split('.')[0]
            attrs[field_name] = DjangoRelationship(
                type_=get_type_name(related_name).lower(),
                many=field['relationship'] == 'HasMany',
                schema=f'{related_name}Schema',
                related_url=f'/forest/{collection_name}/{{{collection_name.lower()}_id}}/relationships/{field_name}',
                related_url_kwargs={f'{collection_name.lower()}_id': '<pk>'},
                id_field='pk'
            )
        else:
            attrs[field_name] = get_marshmallow_field(field, Model)
    return attrs


# override Marshmallow Schema for Django needs
class DjangoSchema(Schema):

    def __init__(self, *args, **kwargs):
        self.include_data = kwargs.pop("include_data", ())
        super(Schema, self).__init__(*args, **kwargs)
        if self.include_data:
            self.check_relations(self.include_data)

        if not self.opts.type_:
            raise ValueError("Must specify type_ class Meta option")

        if self.opts.self_url_kwargs and not self.opts.self_url:
            raise ValueError(
                "Must specify `self_url` Meta option when "
                "`self_url_kwargs` is specified"
            )
        self.included_data = {}
        self.document_meta = {}

    def format_item(self, item):
        ret = super(DjangoSchema, self).format_item(item)
        if ret and self._original:
            ret['id'] = self._original.pk
        return ret

    def format_items(self, data, many):
        if many:
            res = []
            queryset = self._original
            for i, item in enumerate(data):
                self._original = queryset[i]
                res.append(self.format_item(item))
        else:
            res = super(DjangoSchema, self).format_items(data, many)
        return res

    @ma.post_dump(pass_many=True, pass_original=True)
    def format_json_api_response(self, data, original, many):
        self._original = original  # needed to get the id value
        return super(DjangoSchema, self).format_json_api_response(data, many)

    def cast_value(self, field, value):
        return field.get_prep_value(value)

    def get_attribute(self, obj, attr, default):
        value = super().get_attribute(obj, attr, default)
        if value.__class__.__name__ == 'ManyRelatedManager':
            value = value.all()
        else:
            try:
                field = obj.__class__._meta.get_field(attr)
            except FieldDoesNotExist:
                pass
            else:
                if not field.is_relation:
                    value = self.cast_value(field, value)

        return value

    def get_resource_links(self, item):
        id_attr = self._original._meta.pk.attname
        field = self._original._meta.get_field(id_attr)
        item['__id__'] = self.cast_value(field, self._original.pk)
        res = super(DjangoSchema, self).get_resource_links(item)
        del item['__id__']
        return res

def create_json_api_schema(collection):
    collection_name = collection['name']
    attrs = populate_attrs(collection, collection_name)

    class MarshmallowSchema(DjangoSchema):

        class Meta:
            type_ = get_type_name(collection_name).lower()
            self_url = f'/forest/{collection_name}/{{{collection_name.lower()}_id}}'
            self_url_kwargs = {f'{collection_name.lower()}_id': '<__id__>'}
            strict = True

    return MarshmallowType(f'{collection_name}Schema', (MarshmallowSchema,), attrs)



from django.conf import settings
from django.test import TestCase, override_settings
from corsheaders.defaults import default_headers

from django_forest.utils.cors import set_cors


class UtilsCorsTests(TestCase):

    @override_settings(MIDDLEWARE=[], INSTALLED_APPS=[])
    def test_set_cors(self):
        set_cors()
        self.assertTrue('corsheaders' in settings.INSTALLED_APPS)
        self.assertTrue('corsheaders.middleware.CorsMiddleware' in settings.MIDDLEWARE)
        self.assertEqual(settings.MIDDLEWARE.index('django_forest.utils.cors.PnaMiddleware'), 0)
        self.assertEqual(settings.MIDDLEWARE.index('corsheaders.middleware.CorsMiddleware'), 1)
        self.assertEqual(settings.CORS_ALLOWED_ORIGIN_REGEXES, ['.*\\.forestadmin\\.com.*'])
        self.assertEqual(settings.CORS_ALLOW_HEADERS, list(default_headers) + ['forest-context-url'])
        self.assertEqual(settings.CORS_ALLOW_CREDENTIALS, True)

    @override_settings(INSTALLED_APPS=['corsheaders'])
    def test_set_cors_apps_already_present(self):
        set_cors()
        self.assertEqual(settings.INSTALLED_APPS, ['corsheaders'])

    @override_settings(MIDDLEWARE=['fake', 'django.middleware.X', 'corsheaders.middleware.CorsMiddleware'])
    def test_set_cors_middleware_already_present(self):
        set_cors()
        self.assertEqual(settings.MIDDLEWARE, ['fake', 'django.middleware.X', 'django_forest.utils.cors.PnaMiddleware', 'corsheaders.middleware.CorsMiddleware'])
    
    @override_settings(MIDDLEWARE=['fake', 'django.middleware.common.CommonMiddleware', 'corsheaders.middleware.CorsMiddleware'])
    def test_set_cors_with_common_middleware(self):
        self.assertEqual(settings.MIDDLEWARE, ['fake', 'django.middleware.common.CommonMiddleware', 'corsheaders.middleware.CorsMiddleware'])

    @override_settings(MIDDLEWARE=['fakemiddleware', 'django.middleware.common.CommonMiddleware', 'fakemiidleware2'])
    def test_set_cors_common_middleware(self):
        set_cors()
        self.assertTrue('corsheaders.middleware.CorsMiddleware' in settings.MIDDLEWARE)
        self.assertEqual(settings.MIDDLEWARE, ['fakemiddleware', 'django_forest.utils.cors.PnaMiddleware', 'corsheaders.middleware.CorsMiddleware', 'django.middleware.common.CommonMiddleware', 'fakemiidleware2'])

    @override_settings(CORS_ALLOWED_ORIGIN_REGEXES=[r'/\.foo\.com$/',r'/\.bar\.com$/'])
    def test_extends_allowed_origin_regexes(self):
        set_cors()
        self.assertEqual(settings.CORS_ALLOWED_ORIGIN_REGEXES, ['.*\\.forestadmin\\.com.*', '/\\.foo\\.com$/', '/\\.bar\\.com$/'])
    
    @override_settings(CORS_ALLOW_HEADERS=['fake', 'fake2'])
    def test_extends_allowed_headers(self):
        set_cors()
        self.assertEqual(settings.CORS_ALLOW_HEADERS, ['fake', 'fake2', 'forest-context-url'])
    


from django.http import HttpResponse
from ipware import get_client_ip

from django_forest.utils.ip_whitelist import IpWhitelist


class IpWhitelistMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response

    def get_client_ip(self, request):
        client_ip, _ = get_client_ip(request)
        return client_ip

    def is_ip_valid(self, request):
        if IpWhitelist.use_ip_whitelist:
            client_ip = self.get_client_ip(request)
            if client_ip is None:
                return False
            return IpWhitelist.is_ip_matches_any_rule(client_ip)
        return True

    def process_view(self, request, view_func, *args, **kwargs):

        # if invalid ip, fetch again
        if not IpWhitelist.fetched or not self.is_ip_valid(request):
            try:
                IpWhitelist.get_rules()
            except Exception as e:
                return HttpResponse(f'Unable to retrieve the ip white list ({e})', status=403)

        if not self.is_ip_valid(request):
            return HttpResponse('IP client is invalid', status=403)



from django.apps import apps

from django_forest.utils.forest_setting import get_forest_setting


class Models:
    models = None

    @classmethod
    def list(cls, force=False):
        if cls.models is None or force:
            included_models = get_forest_setting('INCLUDED_MODELS')
            excluded_models = get_forest_setting('EXCLUDED_MODELS')
            cls.models = apps.get_models(include_auto_created=True)
            if included_models is not None:
                cls.models = [m for m in cls.models if m._meta.db_table in included_models]
            elif excluded_models is not None:
                cls.models = [m for m in cls.models if m._meta.db_table not in excluded_models]
        return cls.models

    @classmethod
    def get(cls, resource):
        for model in cls.list():
            if resource.lower() in (model._meta.db_table.lower(), f'{model._meta.db_table.lower()}s'):
                return model

        return None



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from . import views

app_name = 'actions'
urlpatterns = [
    path('/send-invoice', csrf_exempt(views.SendInvoiceActionView.as_view()), name='send-invoice'),
    path('/not-exists', views.NotExistsActionView.as_view(), name='not-exists'),
    path('/mark-as-live', csrf_exempt(views.MarkAsLiveActionView.as_view()), name='mark-as-live'),
    path('/generate-invoice', csrf_exempt(views.GenerateInvoiceActionView.as_view()), name='generate-invoice'),
]



import copy
import sys
from datetime import datetime
from unittest import mock
from django.conf import settings
from freezegun import freeze_time

import pytest
import pytz
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.collection import Collection
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.date import get_timezone


# reset forest config dir auto import
from django_forest.utils.scope import ScopeManager


@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class ResourceListViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json',
                'session.json',
                'question.json', 'choice.json',
                'place.json', 'restaurant.json',
                'student.json',
                'serial.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.reverse_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_choice'})
        self.no_data_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_waiter'})
        self.enum_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_student'})
        self.uuid_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_serial'})
        self.one_to_one_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_restaurant'})
        self.bad_url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_foo'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}
        Collection._registry = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_to_edit(self, mocked_datetime):
        settings.MIDDLEWARE.insert(0, 'django_forest.middleware.DeactivateCountMiddleware')
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=pytz.UTC)
        response = self.client.get(self.reverse_url, {
            'context[relationship]': 'HasMany',
            'context[field]': 'choice_set',
            'context[collection]': 'tests_question',
            'context[recordId]': '1',
            'search': 'yes',
            'searchToEdit': 'true'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_choice',
                    'id': 1,
                    'attributes': {
                        'choice_text': 'yes',
                        'votes': 0
                    },
                    'relationships': {
                        'question': {
                            'links': {
                                'related': '/forest/tests_choice/1/relationships/question'
                            },
                            'data': {
                                'type': 'tests_question',
                                'id': '1'
                            }
                        },
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_choice/1/relationships/topic'
                                }
                            }
                    },
                    'links': {
                        'self': '/forest/tests_choice/1'
                    }
                },
            ],
            'included': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'question_text': 'what is your favorite color?',
                        'pub_date': '2021-06-02T13:52:53.528000+00:00'
                    },
                    'id': 1, 'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                },
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                }
            ],
            'meta': {
                'decorators': [
                    {
                        'id': 1,
                        'search': ['choice_text']
                    }
                ]
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_pk_no_id(self, mocked_decode):
        response = self.client.get(self.one_to_one_url, {
            'context[relationship]': 'BelongsTo',
            'context[field]': 'restaurant',
            'context[collection]': 'tests_place',
            'context[recordId]': '1',
            'fields[tests_restaurant]': 'id',
            'search': '1',
            'searchToEdit': 'true'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_restaurant',
                    'id': 1,
                    'relationships': {
                        'place': {
                            'links': {
                                'related': '/forest/tests_restaurant/1/relationships/place'
                            },
                            'data': {
                                'type': 'tests_place',
                                'id': '1'
                            }
                        }
                    },
                    'links': {
                        'self': '/forest/tests_restaurant/1'
                    }
                }
            ],
            'included': [
                {
                    'type': 'tests_place',
                    'relationships': {
                        'restaurant': {
                            'links': {
                                'related': '/forest/tests_place/1/relationships/restaurant'
                            }
                        }
                    },
                    'attributes': {
                        'address': 'Venezia, Italia',
                        'name': 'San Marco'},
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_place/1'
                    }
                }
            ]
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_number(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'page[number]': 1,
            'page[size]': 15,
            'search': 1,
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ],
            'meta': {
                'decorators':[
                    {'id':1, 'search':['id']}
                ]
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_number_max_size(self, mocked_decoded):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date',
            'page[number]': 1,
            'page[size]': 15,
            'search': sys.maxsize + 1,
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': []
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_mutliple(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date,foo,bar',
            'fields[topic]': 'name',
            'page[number]': 1,
            'page[size]': 15,
            'search': 'favorite',
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?',
                        'foo': 'what is your favorite color?+foo',
                        'bar': 'what is your favorite color?+bar',
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?',
                        'foo': 'who is your favorite singer?+foo',
                        'bar': 'who is your favorite singer?+bar',
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ],
            'meta': {
                'decorators': [
                    {
                        'id': 1,
                        'search': ['question_text', 'foo', 'bar']
                    },
                    {
                        'id': 3,
                        'search': ['question_text', 'foo', 'bar']
                    }
                ]
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_insensitive_search_mutliple(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date,foo,bar',
            'fields[topic]': 'name',
            'page[number]': 1,
            'page[size]': 15,
            'search': 'FaVoRite',
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?',
                        'foo': 'what is your favorite color?+foo',
                        'bar': 'what is your favorite color?+bar',
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?',
                        'foo': 'who is your favorite singer?+foo',
                        'bar': 'who is your favorite singer?+bar',
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ],
            'meta': {
                'decorators': [
                    {
                        'id': 1,
                        'search': ['question_text', 'foo', 'bar']
                    },
                    {
                        'id': 3,
                        'search': ['question_text', 'foo', 'bar']
                    }
                ]
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_enum(self, mocked_decode):
        response = self.client.get(self.enum_url, {
            'fields[tests_student]': 'id,year_in_school',
            'page[number]': 1,
            'page[size]': 15,
            'search': 'FR',
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_student',
                    'attributes': {
                        'year_in_school': 'FR'
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_student/1'
                    }
                }
            ],
            'meta': {
                'decorators': [
                    {
                        'id': 1,
                        'search': ['year_in_school']
                    }
                ]
            }
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_search_uuid(self, mocked_datetime):
        response = self.client.get(self.uuid_url, {
            'fields[tests_serial]': 'uuid',
            'page[number]': 1,
            'page[size]': 15,
            'search': '4759e256-a27a-45e1-b248-09fb1523c978',
            'searchExtended': 0
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_serial',
                    'attributes': {
                        'uuid': '4759e256-a27a-45e1-b248-09fb1523c978'
                    },
                    'id': '4759e256-a27a-45e1-b248-09fb1523c978',
                    'links': {
                        'self': '/forest/tests_serial/4759e256-a27a-45e1-b248-09fb1523c978'
                    }
                }
            ],
            'meta': {
                'decorators': [
                    {
                        'id': '4759e256-a27a-45e1-b248-09fb1523c978',
                        'search': ['uuid']
                    }
                ]}
        })

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_get_extended_search(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'page[number]': 1,
            'page[size]': 15,
            'search': 'yes',
            'searchExtended': 1
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'question_text': 'what is your favorite color?',
                        'pub_date': '2021-06-02T13:52:53.528000+00:00'},
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })


    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=get_timezone('UTC'))
    )
    def test_search_on_belongs_to_for_edition(self, mocked_decode):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id',
            'page[number]': 1,
            'page[size]': 15,
            'search': '1',
            'searchExtended': 0,
            'searchToEdit': True
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                }
            ],
            'meta': {
                'decorators':[
                    {'id':1, 'search':['id']}
                ]
            }
        })



import re

from jose import jwt

from django_forest.utils.forest_setting import get_forest_setting


def get_accessor_name(field):
    name = None
    try:
        # model_set
        name = field.get_accessor_name()
    except AttributeError:
        name = field.name
    return name


def get_token(request):
    token = ''
    if 'Authorization' in request.headers:
        token = request.headers['Authorization'].split()[1]
    # NOTICE: Necessary for downloads authentication.
    elif 'cookie' in request.headers:
        REGEX_COOKIE_SESSION_TOKEN = r'forest_session_token=([^;]*)'
        m = re.search(REGEX_COOKIE_SESSION_TOKEN, request.headers['cookie'])
        token = m.group(1)

    auth_secret = get_forest_setting('FOREST_AUTH_SECRET')
    return jwt.decode(token, auth_secret, algorithms=['HS256'])


def get_association_field(Model, association_resource):
    association_field = next((x for x in Model._meta.get_fields()
                              if x.is_relation and get_accessor_name(x) == association_resource), None)
    if association_field is None:
        message = f'cannot find association resource {association_resource} for Model {Model._meta.db_table}'
        raise Exception(message)

    return association_field



import copy
import sys
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


# reset forest config dir auto import
from django_forest.utils.scope import ScopeManager


@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
class ResourceAssociationCsvViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json']

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:associations:csv',
                           kwargs={'resource': 'tests_question', 'pk': 1, 'association_resource': 'choice_set'})
        self.reverse_url = reverse('django_forest:resources:associations:csv',
                                   kwargs={'resource': 'tests_choice', 'pk': 1, 'association_resource': 'question'})
        self.bad_association_url = reverse('django_forest:resources:associations:csv',
                                           kwargs={'resource': 'tests_question', 'pk': 1, 'association_resource': 'foo'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_get(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_choice]': 'id,question,topic,choice_text',
            'fields[question]': 'question_text',
            'fields[topic]': 'name',
            'search': '',
            'searchExtended': '',
            'filename': 'choices',
            'header': 'id,question,choice text',
            'timezone': 'Europe/Paris',
            'page[number]': 10, # csv get should ignore these. If it doesn't we'll only receive the column names, but no rows
            'page[size]': 100
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode('utf-8'), 'id,question,choice text,\r\n1,what is your favorite color?,,yes\r\n2,what is your favorite color?,,no\r\n')

    def test_get_no_association(self, *args, **kwargs):
        response = self.client.get(self.bad_association_url, {
            'fields[tests_choice]': 'id,question,choice_text',
            'fields[question]': 'question_text',
            'search': '',
            'searchExtended': '',
            'filename': 'choices',
            'header': 'id,question,choice text',
            'timezone': 'Europe/Paris'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {'errors': [{'detail': 'cannot find association resource foo for Model tests_question'}]})



import copy
from datetime import datetime
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse
from freezegun import freeze_time

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager
from django_forest.utils.date import get_timezone


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
@mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})    
class ResourceListFilterDateViewTests(TransactionTestCase):
    fixtures = ['question.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @freeze_time(
        lambda: datetime(2021, 6, 2, 17, 0, 0, 0, tzinfo=get_timezone('Europe/Paris'))
    )
    def test_past(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"past","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })


    @freeze_time(
        lambda: datetime(2021, 6, 2, 17, 0, 0, 0, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_future(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"future","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @freeze_time(
        lambda: datetime(2021, 6, 2, 15, 0, 0, 0, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_today(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'filters': '{"field":"pub_date","operator":"today","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @freeze_time(
        lambda: datetime(2021, 6, 2, 17, 53, 0, 0, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_before_x_hours_ago(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"before_x_hours_ago","value":1}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @freeze_time(
        lambda: datetime(2021, 6, 2, 15, 53, 0, 0, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_after_x_hours_ago(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"after_x_hours_ago","value":1}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @freeze_time(
        lambda: datetime(2021, 7, 1, 0, 0, 0, 0, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_previous_quarter(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"previous_quarter","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })

    @freeze_time(
        lambda: datetime(2021, 6, 30, 23, 59, 59, 999999, tzinfo=get_timezone('Europe/Paris')),
    )
    def test_previous_quarter_to_date(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date',
            'fields[topic]': 'name',
            'filters': '{"field":"pub_date","operator":"previous_quarter_to_date","value":null}',
            'timezone': 'Europe/Paris',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'id': 2,
                    'links': {
                        'self': '/forest/tests_question/2'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/2/relationships/topic'
                            }
                        }
                    },
                },
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'id': 3,
                    'links': {
                        'self': '/forest/tests_question/3'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/3/relationships/topic'
                            }
                        }
                    },
                }
            ]
        })



from django.db.models.fields import NOT_PROVIDED

from django_forest.utils.schema.enums import serialize_value


def handle_default_value(field, f):
    if hasattr(field, 'default') and field.default != NOT_PROVIDED and not callable(field.default):
        f['default_value'] = serialize_value(field.default)
    return f



from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.resources.utils.query_parameters import parse_qs
from django_forest.utils.models import Models

class JsonApiSerializerMixin:

    def get_smart_relationships(self, fields):
        return [x['field'] for x in fields
                if x['is_virtual'] and x['reference'] and not isinstance(x['type'], list)]

    def get_include_data(self, fields):
        relationships = [x['field'] for x in fields if x['reference'] and x['relationship'] in ['BelongsTo', 'HasOne']]
        smart_relationships = self.get_smart_relationships(fields)
        return relationships + smart_relationships

    def serialize_field(self, field, parent=None):
        res = field['field']
        if field['is_virtual'] and field['reference']:
            field_name = field['field']
            target_key = field['reference'].split('.')[1]
            res = f'{field_name}.{target_key}'
        if parent:
            res = f'{parent}.{res}'
        return res

    def build_only(self, fields, current_name):
        only = [self.serialize_field(f) for f in fields.pop(current_name)]
        for collection_name, fields in fields.items():
            name = collection_name
            only += [self.serialize_field(f, name) for f in fields]
            only.remove(name)
        return only

    def get_fields_for_collection(self, collection, required_fields=None):
        collection_fields = collection['fields']
        if required_fields:
            collection_fields = list(
                filter(
                    lambda f: f['field'] in required_fields,
                    collection_fields
                )
            )
            Model = Models.get(collection['name'])
            pk_field = Model._meta.pk
            if 'id' in required_fields and pk_field.name != 'id':
                pk_field = next(filter(lambda f: f['field'] == pk_field.name, collection['fields']))
                collection_fields.append(pk_field)
        return collection_fields

    def build_nested_fields(self, field, required_fields):
        if not field['is_virtual']:
            if not isinstance(required_fields, list):
                required_fields = [required_fields]  # need if only one field in the qs

            nested_name = field['reference'].split('.')[0]
            return self.get_fields_for_collection(
                Schema.get_collection(nested_name),
                required_fields
            )

    def get_fields(self, current_collection, qs):
        current_name = current_collection['name']
        qs_fields = qs.get('fields')
        if not qs_fields:
            return {
                f'{current_name}': self.get_fields_for_collection(
                    current_collection
                )
            }
        fields = {
            f"{current_name}": self.get_fields_for_collection(
                current_collection,
                qs_fields.get(current_name),
            )
        }
        for field in fields[current_name]:
            nested_required_fields = qs_fields.get(field['field'])
            if nested_required_fields:
                res = self.build_nested_fields(
                    field,
                    nested_required_fields
                )
                if res:
                    fields[field['field']] = res
        return fields

    def serialize(self, queryset, Model, params):
        db_name = Model._meta.db_table
        current_collection = Schema.get_collection(db_name)
        JsonSchema = JsonApiSchema.get(db_name)
        qs = parse_qs(params)
        fields = self.get_fields(current_collection, qs)
        kwargs = {
            'include_data': self.get_include_data(fields[db_name])
        }

        if db_name in qs.get('fields', {}):
            kwargs['only'] = self.build_only(fields, db_name)

        data = {'data': []}
        if queryset:
            data = JsonSchema(**kwargs).dump(queryset, many=True)
        return data



from django_forest.utils.collection import Collection
from django_forest.utils.schema import Schema


class SmartFieldMixin:
    def _handle_get_method(self, smart_field, item, resource):
        if 'get' in smart_field:
            method = smart_field['get']
            if isinstance(method, str):
                setattr(item, smart_field['field'], getattr(Collection._registry[resource], method)(item))
            elif callable(method):
                setattr(item, smart_field['field'], method(item))

    def _handle_set_method(self, smart_field, instance, value, resource):
        if 'set' in smart_field:
            method = smart_field['set']
            if isinstance(method, str):
                instance = getattr(Collection._registry[resource], method)(instance, value)
            elif callable(method):
                instance = method(instance, value)
        return instance

    def _add_smart_fields(self, item, smart_fields, resource):
        for smart_field in smart_fields:
            self._handle_get_method(smart_field, item, resource)

    def _get_smart_fields_for_request(self, collection, params=None):

        fields = [field for field in collection['fields'] if field['is_virtual']]

        if params is None:
            return fields

        filtered_fields = params.get('fields', {}).get(collection.get('name'))
        if filtered_fields is None:
            return fields

        return [field for field in fields if field['field'] in filtered_fields]

    def handle_smart_fields(self, queryset, resource, params=None, many=False, follow_relations=True):
        collection = Schema.get_collection(resource)

        # Rather than calculate and then filter out smart fields, we want to ignore them entirely
        smart_fields = self._get_smart_fields_for_request(collection, params)

        # Don't bother adding anything if there are no smart fields
        if smart_fields and many:
            for item in queryset:
                self._add_smart_fields(item, smart_fields, resource)
        elif smart_fields:
            self._add_smart_fields(queryset, smart_fields, resource)

        # handle smart fields on nested relations (call handle_smart_fields recursively)
        if follow_relations is True:
            self._handle_smart_field_on_relations(queryset, collection, params, many)

    def _handle_smart_field_on_relations(self, queryset, base_collection, params, many):
        relations = self.__get_relations_for_smart_fields(base_collection, params)
        for relation_field in relations:
            collection_name = relation_field["reference"].split(".")[0]

            transformed_params = None
            if params:
                transformed_params = {
                    "fields": {
                        collection_name: [params.get("fields", {}).get(relation_field["field"])]
                    }
                }
            self._handle_smart_field_for_relation(queryset, relation_field, collection_name, transformed_params, many)

    def _handle_smart_field_for_relation(self, queryset, relation_field, collection_name, params, many):
        if many:
            for item in queryset:
                if getattr(item, relation_field["field"], None) is not None:
                    self.handle_smart_fields(
                        getattr(item, relation_field["field"]), collection_name, params, False, False
                    )
        else:
            if getattr(queryset, relation_field["field"], None) is not None:
                self.handle_smart_fields(
                    getattr(queryset, relation_field["field"]), collection_name, params, False, False
                )

    def __get_relations_for_smart_fields(self, base_collection, params):
        relations = [
            field
            for field in base_collection["fields"]
            if field["relationship"] is not None and field["relationship"] in ["BelongsTo", "HasOne"]
        ]
        # filter with asked fields in params (if set)
        if params is not None:
            relations = [
                rel for rel in relations
                if rel["field"] in params.get("fields", {}).keys()
            ]
        return relations

    def update_smart_fields(self, instance, body, resource):
        collection = Schema.get_collection(resource)
        smart_fields = [x for x in collection['fields'] if x['is_virtual']]
        for smart_field in smart_fields:
            if smart_field['field'] in body['data']['attributes'].keys():
                value = body['data']['attributes'][smart_field['field']]
                instance = self._handle_set_method(smart_field, instance, value, resource)
        return instance



from django.conf import settings


def set_middlewares():
    middlewares = list(settings.MIDDLEWARE)

    if 'django_forest.middleware.PermissionMiddleware' not in middlewares:
        settings.MIDDLEWARE.insert(0, 'django_forest.middleware.PermissionMiddleware')

    if 'django_forest.middleware.IpWhitelistMiddleware' not in middlewares:
        settings.MIDDLEWARE.insert(0, 'django_forest.middleware.IpWhitelistMiddleware')



import json

from django.http import HttpResponse

from django_forest.utils import get_token
from django_forest.utils.permissions import Permission


class PermissionMiddleware:
    mapping = {
        'list': {
            'GET': 'browseEnabled',
            'POST': 'addEnabled',
            'PUT': 'editEnabled',
            'DELETE': 'deleteEnabled'
        },
        'detail': {
            'GET': 'readEnabled',
            'PUT': 'editEnabled',
            'DELETE': 'deleteEnabled'
        },
        'count': {
            'GET': 'browseEnabled'
        },
        'csv': {
            'GET': 'exportEnabled'
        },
        'liveQueries': {
            'POST': 'liveQueries'
        },
        'statsWithParameters': {
            'POST': 'statsWithParameters'
        }
    }

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response

    def populate_query_request_info(self, permission_name, body, kwargs):
        if permission_name in 'liveQueries':
            kwargs['query_request_info'] = body['query']
        elif permission_name in 'statsWithParameters':
            if 'timezone' in body:
                del body['timezone']
            if 'group_by_field' in body:
                #  usefull if the group_by_field is a related field
                body['group_by_field'] = body['group_by_field'].split(':')[0]
            kwargs['query_request_info'] = body

    def is_authorized(self, request, token, resource):
        permission_name = self.mapping[request.resolver_match.url_name][request.method]

        is_chart = permission_name in ('liveQueries', 'statsWithParameters')
        kwargs = {}
        if is_chart:
            body = json.loads(request.body.decode('utf-8'))
            self.populate_query_request_info(permission_name, body, kwargs)

        if not (is_chart and token['permission_level'] in ['admin', 'editor', 'developer']):
            permission = Permission(resource,
                                    permission_name,
                                    token['rendering_id'],
                                    token['id'],
                                    **kwargs)
            if not Permission.is_authorized(permission):
                raise Exception('not authorized')

    def get_resource(self, args):
        resource = None
        if 'resource' in args[1]:
            resource = args[1]['resource']
        return resource

    def process_view(self, request, view_func, *args, **kwargs):
        if request.resolver_match.app_name.startswith('django_forest:resources') or \
                request.resolver_match.app_name.startswith('django_forest:stats'):

            token = get_token(request)
            resource = self.get_resource(args)

            try:
                self.is_authorized(request, token, resource)
            except Exception as e:
                return HttpResponse(f'Unauthorized request ({e})', status=403)



import logging
import csv

from django_forest.resources.utils.csv import CsvMixin
from django_forest.resources.utils.format import FormatFieldMixin
from django_forest.resources.utils.json_api_serializer import JsonApiSerializerMixin
from django_forest.resources.utils.query_parameters import parse_qs
from django_forest.resources.utils.resource import ResourceView
from django_forest.resources.utils.smart_field import SmartFieldMixin

logger = logging.getLogger(__name__)


class CsvView(FormatFieldMixin, SmartFieldMixin, JsonApiSerializerMixin, CsvMixin, ResourceView):
    def get(self, request):
        # default
        queryset = self.Model.objects.all()

        params = request.GET.dict()

        try:
            # enhance queryset, ignoring any parameters about pagination
            queryset = self.enhance_queryset(queryset, self.Model, params, request, apply_pagination=False)

            # handle smart fields
            self.handle_smart_fields(queryset, self.Model._meta.db_table, parse_qs(params), many=True)

            # json api serializer
            data = self.serialize(queryset, self.Model, params)
        except Exception as e:
            logger.exception(e)
            return self.error_response(e)
        else:
            response = self.csv_response(params['filename'])

            field_names_requested = [x for x in params[f'fields[{self.Model._meta.db_table}]'].split(',')]
            csv_header = params['header'].split(',')

            writer = csv.DictWriter(response, fieldnames=field_names_requested)
            writer.writerow(dict(zip(field_names_requested, csv_header)))
            self.fill_csv(data, writer, params)
            return response



from datetime import datetime, timezone

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.timezone import make_aware


class FormatFieldMixin:
    def get_association_instance(self, association, value):
        pk = value['data']['id']
        try:
            instance = association.related_model.objects.get(pk=pk)
        except ObjectDoesNotExist:
            raise Exception(f'Instance {association.related_model._meta.db_table} with pk {pk} does not exists')
        else:
            return instance

    def handle_foreign_key(self, name, value):
        if value['data'] is not None:
            association = next((x for x in self.Model._meta.get_fields() if x.is_relation and x.name == name), None)
            return self.get_association_instance(association, value)
        return None

    def format(self, name, value, field):
        # TODO other special fields
        if isinstance(field, models.DateTimeField):
            return make_aware(datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%fZ'), timezone.utc)
        elif isinstance(field, models.ForeignKey) or isinstance(field, models.OneToOneRel):
            return self.handle_foreign_key(name, value)

        return value

    def get_attributes(self, body, fields):
        attributes = {}
        for k, v in body.items():
            if k in fields.keys():
                attributes[k] = self.format(k, v, fields[k])
        return attributes

    def populate_attribute(self, body):
        fields = {x.name: x for x in self.Model._meta.get_fields() if not x.many_to_many}
        attributes = {}
        for key in ('attributes', 'relationships'):
            if key in body['data']:
                attributes.update(self.get_attributes(body['data'][key], fields))

        return attributes



from django.conf import settings
from django.test import TestCase

from django_forest.utils.middlewares import set_middlewares


class UtilsMiddlewaresTests(TestCase):
    def tearDown(self):
        settings.MIDDLEWARE.remove("django_forest.middleware.PermissionMiddleware")
        settings.MIDDLEWARE.remove("django_forest.middleware.IpWhitelistMiddleware")

    def test_set_middlewares(self):
        set_middlewares()
        self.assertTrue(
            "django_forest.middleware.PermissionMiddleware" in settings.MIDDLEWARE
        )
        self.assertTrue(
            "django_forest.middleware.IpWhitelistMiddleware" in settings.MIDDLEWARE
        )

    def test_set_middlewares_already_present(self):
        settings.MIDDLEWARE.insert(0, "django_forest.middleware.PermissionMiddleware")
        settings.MIDDLEWARE.append("django_forest.middleware.IpWhitelistMiddleware")

        set_middlewares()

        permission_filtered_middlewares = list(
            filter(
                lambda x: x == "django_forest.middleware.PermissionMiddleware",
                settings.MIDDLEWARE,
            )
        )
        ip_whitelist_filtered_middlewares = list(
            filter(
                lambda x: x == "django_forest.middleware.IpWhitelistMiddleware",
                settings.MIDDLEWARE,
            )
        )
        # exists only once
        self.assertEqual(len(permission_filtered_middlewares), 1)
        self.assertEqual(len(ip_whitelist_filtered_middlewares), 1)

        # order preserved
        self.assertEqual(settings.MIDDLEWARE[0], "django_forest.middleware.PermissionMiddleware")
        self.assertEqual(settings.MIDDLEWARE[-1], "django_forest.middleware.IpWhitelistMiddleware")


import datetime
import uuid

from django.contrib.postgres.fields import ArrayField
from django.core.validators import MinLengthValidator
from django.db import models


# ForeignKey example
from django.db.models import UUIDField


class Topic(models.Model):
    name = models.CharField(max_length=200)

class Question(models.Model):
    question_text = models.CharField(max_length=200, validators=[MinLengthValidator(1)], default='how are you?')
    pub_date = models.DateTimeField('date published', default=datetime.datetime(2020, 5, 17))
    topic = models.ForeignKey(Topic, on_delete=models.CASCADE, null=True)

    def __str__(self):
        return self.question_text


class Choice(models.Model):
    question = models.ForeignKey(Question, on_delete=models.CASCADE, null=True)
    choice_text = models.CharField(max_length=200)
    votes = models.IntegerField(default=0)

    def __str__(self):
        return self.choice_text


# Complex ForeignKey example
class Car(models.Model):
    brand = models.CharField(max_length=200, unique=True)


class Wheel(models.Model):
    car = models.ForeignKey(Car, on_delete=models.CASCADE, related_name='wheels', to_field='brand')


# custom pk example
class Session(models.Model):
    session_key = models.CharField(max_length=40, primary_key=True)
    session_data = models.TextField()
    expire_date = models.DateTimeField()


# ManyToManyField example
class Publication(models.Model):
    title = models.CharField(max_length=30)

    def __str__(self):
        return self.title


class Article(models.Model):
    headline = models.CharField(max_length=100)
    publications = models.ManyToManyField(Publication)

    def __str__(self):
        return self.headline


# OneToOneField example
class Place(models.Model):
    name = models.CharField(max_length=50)
    address = models.CharField(max_length=80)

    def __str__(self):
        return "%s the place" % self.name


class Restaurant(models.Model):
    place = models.OneToOneField(
        Place,
        on_delete=models.CASCADE,
        primary_key=True,
    )
    serves_hot_dogs = models.BooleanField(default=False)
    serves_pizza = models.BooleanField(default=False)

    def __str__(self):
        return "%s the restaurant" % self.place.name


class Waiter(models.Model):
    restaurant = models.ForeignKey(Restaurant, on_delete=models.CASCADE)
    name = models.CharField(max_length=50)

    def __str__(self):
        return "%s the waiter at %s" % (self.name, self.restaurant)


# Enums example
class Student(models.Model):
    FRESHMAN = 'FR'
    SOPHOMORE = 'SO'
    JUNIOR = 'JR'
    SENIOR = 'SR'
    GRADUATE = 'GR'
    YEAR_IN_SCHOOL_CHOICES = [
        (FRESHMAN, 'Freshman'),
        (SOPHOMORE, 'Sophomore'),
        (JUNIOR, 'Junior'),
        (SENIOR, 'Senior'),
        (GRADUATE, 'Graduate'),
    ]
    year_in_school = models.CharField(
        max_length=2,
        choices=YEAR_IN_SCHOOL_CHOICES,
        default=FRESHMAN,
    )


# ArrayField example
class ChessBoard(models.Model):
    board = ArrayField(
        models.CharField(max_length=10, blank=True),
        size=8,
    )


# UUID Field
class Serial(models.Model):
    uuid = UUIDField(primary_key=True, default=uuid.uuid4, editable=False)



from django.http import JsonResponse
from django_forest.authentication.exception import BaseAuthenticationException, AuthenticationOpenIdClientException


def authentication_exception(f):
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except AuthenticationOpenIdClientException as error:
            return JsonResponse(
                {
                    "error": error.error,
                    "error_description": error.error_description,
                    "state": error.state
                },
                status=error.STATUS
            )

        except BaseAuthenticationException as error:
            return JsonResponse({'errors': [{'detail': str(error)}]}, status=error.STATUS)
    return wrapper



import json

from django_forest.resources.utils.resource import ResourceView
from django_forest.stats.utils.stats import StatsMixin
from django_forest.utils import get_association_field

from .utils import get_annotated_queryset, get_format_time_frame, compute_value, compute_line_values, get_periods, \
    contains_previous_date_operator


class StatsWithParametersView(StatsMixin, ResourceView):
    def compute_data(self, key, value, queryset):
        data = {}
        for x in queryset:
            if x[value] is not None:
                self.fill_data(data, x[key], int(x[value]))
        return [{
            'key': k,
            'value': v
        } for k, v, in data.items()]

    def get_previous_count(self, params, request):
        self.previous = True
        queryset = self.enhance_queryset(self.Model.objects.all(), self.Model, params, request)
        return compute_value(params, queryset)

    def handle_count_previous(self, params, res, request):
        if params['type'] == 'Value' and 'filters' in params:
            filters = json.loads(params['filters'])
            if contains_previous_date_operator(filters):
                res['countPrevious'] = self.get_previous_count(params, request)

    def get_value(self, params, queryset, request):
        value = compute_value(params, queryset)
        key = 'countCurrent'
        if params['type'] == 'Objective':
            key = 'value'

        res = {
            key: value
        }

        # Notice: handle countPrevious
        self.handle_count_previous(params, res, request)

        return res

    def get_pie(self, params, queryset):
        pk_name = self.Model._meta.pk.name
        group_by_field = params['group_by_field'].replace(":", "__")

        queryset = queryset.order_by(group_by_field).values(group_by_field)

        aggregate = params['aggregate'].lower()
        queryset, name = get_annotated_queryset(params, queryset, pk_name)

        return self.compute_data(group_by_field, f'{name}__{aggregate}', queryset)

    def get_line(self, params, queryset):
        pk_name = self.Model._meta.pk.name
        format, time_frame = get_format_time_frame(params)

        group_by_date_field = params['group_by_date_field']
        queryset = queryset.order_by(group_by_date_field).values(group_by_date_field)
        aggregate = params['aggregate'].lower()
        queryset, name = get_annotated_queryset(params, queryset, pk_name)

        try:
            bounds = {
                'earliest': queryset.first().get(group_by_date_field, None),
                'latest': queryset.last().get(group_by_date_field, None)
            }
        except Exception:
            values = []
        else:
            periods = get_periods(queryset, f'{name}__{aggregate}', group_by_date_field, format)
            values = compute_line_values(bounds, periods, time_frame, format)
        return values

    def get_leaderboard(self, params, queryset):
        label_field = params['label_field']
        relationship_field = params['relationship_field']

        queryset = queryset.values(label_field)
        association_field = get_association_field(self.Model, relationship_field)
        pk_name = association_field.related_model._meta.pk.name

        aggregate = params['aggregate'].lower()
        association_field_name = association_field.name

        queryset, name = get_annotated_queryset(params, queryset, pk_name, association_field_name)
        queryset = queryset.order_by(f'-{name}__{aggregate}')
        queryset = queryset[:params['limit']]

        return self.compute_data(label_field, f'{name}__{aggregate}', queryset)

    def post(self, request, *args, **kwargs):
        params = self.get_body(request.body)
        params.update(request.GET.dict())
        queryset = self.Model.objects.all()
        queryset = self.enhance_queryset(queryset, self.Model, params, request)
        return self.chart(params, request, queryset)



import os
from distutils.util import strtobool

from django.conf import settings


def get_forest_setting(setting, default=None):
    django_settings = getattr(settings, 'FOREST', {}).get(setting, default)
    forest_setting = os.getenv(setting, django_settings)
    if isinstance(default, bool) and isinstance(forest_setting, str):
        try:
            forest_setting = strtobool(forest_setting)
        except ValueError:
            forest_setting = default
    return forest_setting



APIMAP_ERRORS = {
    404: {
        'level': 'error',
        'message': 'Cannot find the project related to the envSecret you configured. Can you check on Forest that you copied it properly in the Forest settings?'  # noqa E501
    },
    503: {
        'level': 'warning',
        'message': 'Forest is in maintenance for a few minutes. We are upgrading your experience in the forest. We just need a few more minutes to get it right.'  # noqa E501
    },
    'error': {
        'level': 'error',
        'message': 'An error occured with the apimap sent to Forest. Please contact support@forestadmin.com for further investigations.'  # noqa E501
    }
}



import copy
import json

from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.scope import ScopeManager


class StatsLiveQueriesViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json',]

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.handle_json_api_schema()
        self.url = f"{reverse('django_forest:stats:liveQueries')}?timezone=Europe%2FParis"

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    def test_get_value(self):
        body = {
            'query': '''SELECT COUNT(*) as value
                        FROM tests_question''',
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'countCurrent': 3
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_value_wrong(self):
        body = {
            'query': '''SELECT COUNT(*) as id__count
                        FROM tests_question''',
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': "The result column must be named 'value' instead of 'id__count'."
                }
            ]
        })

    def test_get_value_wrong_parameters(self):
        body = {
            'query': '''SELECT COUNT(*) as key, 5 as value
                        FROM tests_question''',
            'type': 'Value'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': "The result column must be named 'value'"
                }
            ]
        })

    def test_get_objective(self):
        body = {
            'query': '''SELECT COUNT(*) as value, 5 as objective
                        FROM tests_question''',
            'type': 'Objective'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': {
                'value': 3,
                'objective': 5
            }
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_objective_wrong(self):
        body = {
            'query': '''SELECT COUNT(*) as value
                        FROM tests_question''',
            'type': 'Objective'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': "The result columns must be named 'value' and 'objective'."
                }
            ]
        })

    def test_get_objective_wrong_parameters(self):
        body = {
            'query': '''SELECT COUNT(*) as key, 5 as value
                        FROM tests_question''',
            'type': 'Objective'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': "The result columns must be named 'value', 'objective' instead of 'key', 'value'."
                }
            ]
        })

    def test_get_pie(self):
        body = {
            'query': '''SELECT tests_question.question_text as key,
                          COUNT(tests_question.id) AS value
                          FROM tests_question
                          WHERE tests_question.pub_date IS NOT NULL
                          GROUP BY tests_question.question_text
                          ORDER BY tests_question.question_text ASC''',
            'type': 'Pie'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': 'do you like chocolate?',
                    'value': 1
                },
                {
                    'key': 'what is your favorite color?',
                    'value': 1
                },
                {
                    'key': 'who is your favorite singer?',
                    'value': 1
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line(self):
        body = {
            'query': '''SELECT tests_question.pub_date as key,
                        COUNT(tests_question.id) AS value
                        FROM tests_question
                        GROUP BY tests_question.pub_date 
                        ORDER BY tests_question.pub_date
                        ASC''',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': '02/06/2021',
                    'values': {
                        'value': 2
                    }
                },
                {
                    'label': '03/06/2021',
                    'values': {
                        'value': 1
                    }
                },
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_line_date(self):
        body = {
            'query': '''SELECT tests_question.pub_date as key,
                        COUNT(tests_question.id) AS value
                        FROM tests_question
                        GROUP BY tests_question.pub_date
                        ORDER BY tests_question.pub_date
                        ASC''',
            'type': 'Line'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'label': '02/06/2021',
                    'values': {
                        'value': 2
                    }
                },
                {
                    'label': '03/06/2021',
                    'values': {
                        'value': 1
                    }
                },
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')

    def test_get_leaderboard(self):
        body = {
            'query': '''SELECT tests_question.question_text as key,
                        COUNT(tests_choice.id) AS value
                        FROM tests_question
                        LEFT OUTER JOIN tests_choice ON (tests_question.id = tests_choice.question_id)
                        GROUP BY tests_question.question_text
                        ORDER BY value
                        DESC LIMIT 5''',
            'type': 'Leaderboard'
        }

        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['data']['attributes'], {
            'value': [
                {
                    'key': 'what is your favorite color?',
                    'value': 2
                },
                {
                    'key': 'do you like chocolate?',
                    'value': 1
                },
                {
                    'key': 'who is your favorite singer?',
                    'value': 0
                }
            ]
        })
        self.assertEqual(data['data']['type'], 'stats')



from django_forest.middleware.ip_whitelist import IpWhitelistMiddleware
from django_forest.middleware.permissions import PermissionMiddleware
from django_forest.middleware.deactivate_count import DeactivateCountMiddleware

__all__ = ['PermissionMiddleware', 'IpWhitelistMiddleware', 'DeactivateCountMiddleware']



import logging

from django.http import JsonResponse, HttpResponse

from django_forest.utils import get_token
from django_forest.utils.collection import Collection
from django_forest.utils.views.base import BaseView

logger = logging.getLogger(__name__)


class HookView(BaseView):

    def action_name_from_endpoint(self, action):
        return str(action['endpoint'].split('/')[-1])

    def hook_exists(self, action, action_name, hook):
        return self.action_name_from_endpoint(action) == action_name and 'hooks' in action and hook in action['hooks']

    def value_in_enums(self, field):
        return next((value for value in field['value'] if value in field['enums']), None)

    def handle_enums(self, field):
        if isinstance(field['type'], list):
            if isinstance(field['value'], list) and self.value_in_enums(field) is None:
                field['value'] = None
        elif field['value'] not in field['enums']:
            field['value'] = None

    def format_enums(self, data):
        for field in data:
            if 'enums' in field and isinstance(field['enums'], list):
                self.handle_enums(field)

    def _get_action(self, action_name):
        for collection_schema in Collection._registry.values():
            for action in collection_schema.actions:
                if self.action_name_from_endpoint(action) == action_name:
                    return action

        raise Exception('action not found')

    def get_hook_result(self, action_name, request):
        action = self._get_action(action_name)
        data = self.handle_action(action, action_name, request)
        self.format_enums(data)
        return data

    def post(self, request, action_name, *args, **kwargs):
        try:
            get_token(request)
        except Exception:
            return HttpResponse(status=403)
        else:
            try:
                data = self.get_hook_result(action_name, request)
            except Exception as e:
                logger.exception(e)
                return self.error_response(e)
            else:
                return JsonResponse({'fields': data}, safe=False)



from unittest import mock

from django.test import TestCase

from django_forest.authentication.oidc.configuration_retriever import retrieve
from django_forest.tests.utils.test_forest_api_requester import mocked_requests

mocked_config = {
    'authorization_endpoint': 'https://api.development.forestadmin.com/oidc/auth',
    'device_authorization_endpoint': 'https://api.development.forestadmin.com/oidc/device/auth',
    'claims_parameter_supported': False,
    'claims_supported': ['sub', 'email', 'sid', 'auth_time', 'iss'],
    'code_challenge_methods_supported': ['S256'],
    'end_session_endpoint': 'https://api.development.forestadmin.com/oidc/session/end',
    'grant_types_supported': ['authorization_code', 'urn:ietf:params:oauth:grant-type:device_code'],
    'id_token_signing_alg_values_supported': ['HS256', 'RS256'],
    'issuer': 'https://api.development.forestadmin.com',
    'jwks_uri': 'https://api.development.forestadmin.com/oidc/jwks',
    'registration_endpoint': 'https://api.development.forestadmin.com/oidc/reg',
    'response_modes_supported': ['query'],
    'response_types_supported': ['code', 'none'],
    'scopes_supported': ['openid', 'email', 'profile'],
    'subject_types_supported': ['public'],
    'token_endpoint_auth_methods_supported': ['none'],
    'token_endpoint_auth_signing_alg_values_supported': [],
    'token_endpoint': 'https://api.development.forestadmin.com/oidc/token',
    'request_object_signing_alg_values_supported': ['HS256', 'RS256'],
    'request_parameter_supported': False,
    'request_uri_parameter_supported': True,
    'require_request_uri_registration': True,
    'claim_types_supported': ['normal']
}


class AuthenticationOidcConfigurationRetrieverTests(TestCase):

    @mock.patch('requests.get', return_value=mocked_requests(mocked_config, 200))
    def test_retrieve(self, mocked_requests_get):
        configuration = retrieve()
        self.assertEqual(configuration, mocked_config)

    @mock.patch('requests.get', return_value=mocked_requests({'error': 'error'}, 400))
    def test_retrieve_error(self, mocked_requests_get):
        with self.assertRaises(Exception) as cm:
            retrieve()

        self.assertEqual(cm.exception.args[0], 'Failed to retrieve the provider\'s configuration.')



import re
from django.conf import settings
from corsheaders.defaults import default_headers

from django_forest.utils.forest_setting import get_forest_setting


#  Waiting cors-header implem
def PnaMiddleware(get_response):
    def middleware(request):
        response = get_response(request)
        if request.headers.get('Access-Control-Request-Private-Network'):
            res = any([
                re.match(pattern, request.headers['origin']) for pattern in settings.CORS_ALLOWED_ORIGIN_REGEXES
            ])
            if res:
                response['Access-Control-Allow-Private-Network'] = 'true'
        return response

    return middleware


def get_list_setting(setting):
    return [x for x in get_forest_setting(setting, '').split(',') if x != '']


def set_cors():

    settings.INSTALLED_APPS = list(settings.INSTALLED_APPS)
    if 'corsheaders' not in settings.INSTALLED_APPS:
        settings.INSTALLED_APPS += ['corsheaders']

    settings.MIDDLEWARE = list(settings.MIDDLEWARE)
    if 'corsheaders.middleware.CorsMiddleware' not in settings.MIDDLEWARE:
        try:
            common_middleware_index = settings.MIDDLEWARE.index('django.middleware.common.CommonMiddleware')
        except ValueError:
            common_middleware_index = 0
        settings.MIDDLEWARE.insert(common_middleware_index, 'corsheaders.middleware.CorsMiddleware')

    pna_index = settings.MIDDLEWARE.index('corsheaders.middleware.CorsMiddleware')
    settings.MIDDLEWARE.insert(pna_index, 'django_forest.utils.cors.PnaMiddleware')

    settings.CORS_ALLOWED_ORIGIN_REGEXES = [
        r'.*\.forestadmin\.com.*',
        *getattr(settings, 'CORS_ALLOWED_ORIGIN_REGEXES', [])
    ]

    settings.CORS_ALLOW_HEADERS = [
        *getattr(settings, 'CORS_ALLOW_HEADERS', list(default_headers)),
        'forest-context-url',
    ]

    settings.CORS_ALLOW_CREDENTIALS = True



from django_forest.tests.forest.choice import ChoiceForest
from django_forest.tests.forest.question import QuestionForest
from django_forest.tests.forest.place import PlaceForest

__all__ = ['QuestionForest', 'ChoiceForest', 'PlaceForest']



from datetime import datetime, timedelta
from django.db.models import Q

from django_forest.resources.utils.queryset.filters.date import DatesMixin
from django_forest.resources.utils.queryset.filters.date.factory import ConditionFactory as DateConditionFactory
from django_forest.utils import get_association_field
from django_forest.utils.type_mapping import get_type
from django_forest.utils.collection import Collection
from django_forest.utils.schema import Schema


OPERATORS = {
    'not': '',
    'contains': '__contains',
    'not_contains': '__contains',
    'before': '__lt',
    'less_than': '__lt',
    'after': '__gt',
    'greater_than': '__gt',
    'starts_with': '__startswith',
    'ends_with': '__endswith',
    'not_equal': '',
    'equal': '',
    'includes_all': '__contains',
    'in': '__in',
}

INSENSITIVE_OPERATORS = {
    'not': '__iexact',
    'contains': '__icontains',
    'not_contains': '__icontains',
    'before': '__lt',
    'less_than': '__lt',
    'after': '__gt',
    'greater_than': '__gt',
    'starts_with': '__istartswith',
    'ends_with': '__iendswith',
    'not_equal': '__iexact',
    'equal': '__iexact',
    'includes_all': '__icontains',
    'in': '__in',
}


class ConditionsMixin(DatesMixin):
    DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
    DATE_FMT = "%Y-%m-%d"

    def get_basic_expression(self, field, field_type, operator, value):
        operators = INSENSITIVE_OPERATORS if isinstance(value, str) else OPERATORS

        try:
            lookup_field = f"{field}{operators[operator]}"
        except Exception:
            raise Exception(f'Unknown provided operator {operator}')
        else:
            is_negated = operator.startswith('not')

            if field_type == 'Date' and operator in ['equal', 'not_equal']:
                kwargs = {f'{lookup_field}__range': (value, value + timedelta(seconds=1))}
            else:
                kwargs = {lookup_field: value}

            if is_negated:
                return ~Q(**kwargs)

            return Q(**kwargs)

    def handle_blank(self, field_type, field):
        if field_type == 'String':
            return Q(Q(**{f'{field}__isnull': True}) | Q(**{f'{field}__exact': ''}))
        return Q(**{f'{field}__isnull': True})

    def get_expression_smart_field(self, smart_fields, condition, resource):
        operator = condition['operator']
        splitted_fields = condition['field'].split(":")
        value = condition['value']

        smart_field = smart_fields[splitted_fields[0]]
        if 'filter' in smart_field:
            method = smart_field['filter']
            if isinstance(method, str):
                ret = getattr(Collection._registry[resource], method)(operator, value)
            elif callable(method):
                ret = method(operator, value)

            if isinstance(ret, Q) and len(splitted_fields) > 1:
                ret = self._rewrite_Q_with_field(ret, condition)

            return ret

    def _rewrite_Q_with_field(self, q, condition):
        operator = condition['operator']
        splitted_fields = condition['field'].split(":")

        new_children = []
        for children in q.children:
            if isinstance(children, Q):
                new_children.append(self._rewrite_Q_with_field(children, condition))
            else:
                if OPERATORS[operator] != "" and OPERATORS[operator] in children[0]:
                    str_qs = children[0].replace(
                        OPERATORS[operator],
                        f'__{"__".join(splitted_fields[1:])}{OPERATORS[operator]}'
                    )
                else:
                    str_qs = f'{"__".join([children[0], *splitted_fields[1:]])}'
                new_children.append((str_qs, children[1]))
        q.children = new_children
        return q

    def _cast_dateonly_field(self, value):
        try:
            value = datetime.strptime(value, self.DATETIME_FMT).date()
        except ValueError:
            value = datetime.strptime(value, self.DATE_FMT).date()
        return value

    def cast_date_fields(self, value, field_type):
        if field_type == 'Date':
            value = datetime.strptime(value, self.DATETIME_FMT)
        elif field_type == 'Dateonly':
            value = self._cast_dateonly_field(value)
        return value

    def get_expression_field(self, condition, Model, tz):
        operator = condition['operator']
        field = condition['field'].replace(':', '__')
        value = condition['value']

        field_type = self.get_field_type(condition['field'], Model)
        if field_type in ['Date', 'Dateonly'] and value and operator not in [
            'previous_x_days',
            'previous_x_days_to_date',
            'before_x_hours_ago',
            'after_x_hours_ago'
        ]:
            value = self.cast_date_fields(value, field_type)

        # special case date, blank and present
        if operator in DateConditionFactory.OPERATORS:
            return self.handle_date_operator(operator, field, value, tz)
        if operator == 'blank':
            return self.handle_blank(field_type, field)
        elif operator == 'present':
            return Q(**{f'{field}__isnull': False})
        else:
            return self.get_basic_expression(field, field_type, operator, value)

    def get_expression(self, condition, Model, tz):
        field = condition['field'].replace(':', '__')

        resource = Model._meta.db_table
        collection = Schema.get_collection(resource)
        smart_fields = {x['field']: x for x in collection['fields'] if x['is_virtual']}
        if field.split("__")[0] in smart_fields.keys():
            return self.get_expression_smart_field(smart_fields, condition, resource)
        else:
            return self.get_expression_field(condition, Model, tz)

    def handle_aggregator(self, filters, Model, tz):
        q_objects = Q()
        for condition in filters['conditions']:
            if "aggregator" in condition:
                if condition['aggregator'] == 'or':
                    q_objects |= self.handle_aggregator(condition, Model, tz)
                else:
                    q_objects &= self.handle_aggregator(condition, Model, tz)
            else:
                if filters['aggregator'] == 'or':
                    q_objects |= self.get_expression(condition, Model, tz)
                else:
                    q_objects &= self.get_expression(condition, Model, tz)
        return q_objects

    def get_field_type(self, field, Model):
        if ':' in field:
            fields = field.split(':')
            RelatedModel = get_association_field(Model, fields[0]).related_model
            field_type = get_type(RelatedModel._meta.get_field(fields[1]))
        else:
            field_type = get_type(Model._meta.get_field(field))

        return field_type



from django_forest.resources.utils.resource import ResourceView


class CountView(ResourceView):
    def get(self, request):
        queryset = self.Model.objects.all()
        params = request.GET.dict()
        return self.get_count(queryset, params, request)



import json
import sys
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class ChangeHooksViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json', ]

    def setUp(self):
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:actions:hooks:change', kwargs={'action_name': 'send-invoice'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_change(self, mocked_decode):
        body = {
            'fields': [
                {
                    'field': 'country',
                    'type': 'Enum',
                    'reference': None,
                    'enums': [
                        'FR',
                        'US'
                    ],
                    'description': None,
                    'isRequired': None,
                    'value': 'FR',
                    'previousValue': 'FR',
                    'widgetEdit': {
                        'name': 'dropdown',
                        'parameters': {
                            'isSearchable': False
                        }
                    },
                    'isReadOnly': None
                },
                {
                    'field': 'city',
                    'type': 'String',
                    'reference': None,
                    'enums': None,
                    'description': None,
                    'isRequired': None,
                    'value': 'foo',
                    'previousValue': None,
                    'widgetEdit': {
                        'name': 'text editor',
                        'parameters': {}
                    },
                    'isReadOnly': None,
                    'hook': 'cityChange'
                },
                {
                    'field': 'zip code',
                    'type': 'String',
                    'reference': None,
                    'enums': None,
                    'description': None,
                    'isRequired': None,
                    'value': None,
                    'previousValue': None,
                    'widgetEdit': {
                        'name': 'text editor',
                        'parameters': {}
                    },
                    'isReadOnly': None,
                    'hook': 'zipCodeChange'
                }
            ],
            'recordIds': [
                '70'
            ],
            'collectionName': 'tests_question',
            'changedField': 'city'
        }
        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data, {
            'fields': [
                {'field': 'country', 'type': 'Enum', 'reference': None, 'enums': ['FR', 'US'], 'description': None,
                 'isRequired': None, 'value': 'FR', 'previousValue': 'FR',
                 'widgetEdit': {'name': 'dropdown', 'parameters': {'isSearchable': False}}, 'isReadOnly': None},
                {'field': 'city', 'type': 'String', 'reference': None, 'enums': None, 'description': None,
                 'isRequired': None, 'value': 'foo', 'previousValue': None,
                 'widgetEdit': {'name': 'text editor', 'parameters': {}}, 'isReadOnly': None, 'hook': 'cityChange'},
                {'field': 'zip code', 'type': 'String', 'reference': None, 'enums': None, 'description': None,
                 'isRequired': None, 'value': '83', 'previousValue': None,
                 'widgetEdit': {'name': 'text editor', 'parameters': {}}, 'isReadOnly': None,
                 'hook': 'zipCodeChange'}]
        })



from django.test import TestCase

from django_forest.tests.models import Question
from django_forest.utils.models import Models


class UtilsGetModelTests(TestCase):
    def test_get_model(self):
        Model = Models.get('tests_question')
        self.assertEqual(Model, Question)

    def test_get_model_None(self):
        Model = Models.get('tests_foo')
        self.assertEqual(Model, None)



from .mark_as_live import MarkAsLiveActionView
from .send_invoice import SendInvoiceActionView
from .not_exists import NotExistsActionView
from .generate_invoice import GenerateInvoiceActionView

__all__ = ['SendInvoiceActionView', 'NotExistsActionView', 'GenerateInvoiceActionView', 'MarkAsLiveActionView']



import json


# TODO handle other special type?
def serialize_value(value):
    try:
        json.dumps(value)
    except TypeError:
        value = str(value)
    return value


def handle_enums(field, f):
    # TODO handle array of enums?
    if f['type'] == 'Enum':
        choices = field.get_choices() if field.blank else field.choices
        f['enums'] = [serialize_value(x[0]) for x in choices]
    return f



import os
import pytest
from django.test import TestCase, override_settings

from django_forest.utils.forest_setting import get_forest_setting


class UtilsCorsTests(TestCase):
    
    @override_settings(
        FOREST={
            'FOO': 'foo',
            'BAR': 'bar'
    })
    def test_get_forest_setting(self):
        self.assertEqual(
            get_forest_setting('FOO'), 
            'foo',
        )
        self.assertEqual(
            get_forest_setting('BAR'), 
            'bar',
        )
        self.assertEqual(
            get_forest_setting('UNKNOWN'), 
            None,
        )
    
    @override_settings(
        FOREST={
            'FOO': 'foo',
            'BAR': 'False'
    })
    def test_get_forest_setting_default(self):
        self.assertEqual(
            get_forest_setting('FOO', 'default'), 
            'foo'
        )
        self.assertEqual(
            get_forest_setting('FOO', True), 
            True,
        )
        self.assertEqual(
            get_forest_setting('BAR', True), 
            False
        )
        self.assertEqual(
            get_forest_setting('Unknown', True), 
            True
        )

    @override_settings(
        FOREST={
            'FOO': 'foo',
            'BAR': 'bar'
    })
    def test_get_forest_setting_env(self):
        os.environ['FOO'] = 'envfoo'
        self.assertEqual(
            get_forest_setting('FOO'), 
            'envfoo'
        )
        self.assertEqual(
            get_forest_setting('BAR'), 
            'bar'
        )
        os.environ['BOOL'] = 'False'
        self.assertEqual(
            get_forest_setting('BOOL', True), 
            False
        )




import logging
import requests
from django_forest.authentication.exception import AuthenticationThirdPartyException

from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester

logger = logging.getLogger(__name__)


def retrieve():
    url = ForestApiRequester.build_url('/oidc/.well-known/openid-configuration')
    try:
        response = ForestApiRequester.get(url)
    except requests.exceptions.RequestException:
        raise AuthenticationThirdPartyException(
            MESSAGES['SERVER_TRANSACTION']['OIDC_CONFIGURATION_RETRIEVAL_FAILED']
        )
    else:
        if response.status_code == requests.codes.ok:
            logger.info(
                f"Retrieve authentication configuration successfull \n {response.json()}"
            )
            return response.json()
        else:
            logger.warning(
                f"Unable to retrive the authentication configuration \n {response.json()}"
            )
            raise AuthenticationThirdPartyException(
                MESSAGES['SERVER_TRANSACTION']['OIDC_CONFIGURATION_RETRIEVAL_FAILED']
            )



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from . import views

app_name = 'associations'
urlpatterns = [
    path('/<slug:association_resource>', csrf_exempt(views.ListView.as_view()), name='list'),
    path('/<slug:association_resource>.csv', csrf_exempt(views.CsvView.as_view()), name='csv'),
    path('/<slug:association_resource>/count', views.CountView.as_view(), name='count'),
]



import sys
from datetime import datetime
from unittest import mock

import pytest
import pytz
from django.test import TransactionTestCase
from django.urls import reverse
from freezegun import freeze_time

from django_forest.tests.resources.views.list.test_list_scope import mocked_scope
from django_forest.utils.collection import Collection
from django_forest.utils.models import Models
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.date import get_timezone


# reset forest config dir auto import
from django_forest.utils.scope import ScopeManager


@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class ResourceListSmartRelationshipViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_choice'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        Collection._registry = {}
        ScopeManager.cache = {}
        Schema.schema_data = None
        Models.models = None

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @freeze_time(
        lambda: datetime(2021, 6, 4, 9, 20, 22, 582772, tzinfo=get_timezone('UTC'))
    )
    @mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
    def test_get(self, mocked_scope_has_expired, mocked_decode):
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }
        response = self.client.get(self.url, {
            'fields[tests_choice]': 'id,question,topic,choice_text,votes',
            'fields[question]': 'question_text',
            'fields[topic]': 'name',
            'page[number]': '1',
            'page[size]': '15'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [{'type': 'tests_choice', 'attributes': {'choice_text': 'yes', 'votes': 0}, 'relationships': {
                'topic': {'links': {'related': '/forest/tests_choice/1/relationships/topic'}, 'data': None},
                'question': {'links': {'related': '/forest/tests_choice/1/relationships/question'},
                             'data': {'type': 'tests_question', 'id': '1'}}}, 'id': 1, 'links': {'self': '/forest/tests_choice/1'}},
                     {'type': 'tests_choice', 'attributes': {'choice_text': 'no', 'votes': 1}, 'relationships': {
                         'topic': {'links': {'related': '/forest/tests_choice/2/relationships/topic'}, 'data': None},
                         'question': {'links': {'related': '/forest/tests_choice/2/relationships/question'},
                                      'data': {'type': 'tests_question', 'id': '1'}}}, 'id': 2,
                      'links': {'self': '/forest/tests_choice/2'}},
                     {'type': 'tests_choice', 'attributes': {'choice_text': 'good', 'votes': 1}, 'relationships': {
                         'topic': {'links': {'related': '/forest/tests_choice/3/relationships/topic'}, 'data': None},
                         'question': {'links': {'related': '/forest/tests_choice/3/relationships/question'},
                                      'data': {'type': 'tests_question', 'id': '2'}}}, 'id': 3,
                      'links': {'self': '/forest/tests_choice/3'}}], 'included': [
                {'type': 'tests_question', 'attributes': {'question_text': 'what is your favorite color?'}, 'id': 1,
                 'links': {'self': '/forest/tests_question/1'}},
                {'type': 'tests_question', 'attributes': {'question_text': 'do you like chocolate?'}, 'id': 2,
                 'links': {'self': '/forest/tests_question/2'}}]})



from datetime import datetime

from django.http import HttpResponse

from django_forest.utils.views.action import ActionView


class GenerateInvoiceActionView(ActionView):

    def post(self, request, *args, **kwargs):
        params = request.GET.dict()
        body = self.get_body(request.body)
        ids = self.get_ids_from_request(request, self.Model)

        file_data = 'foo'
        return HttpResponse(
            file_data,
            content_type='text/txt; charset=utf-8',
            headers={
                'Content-Disposition': f'attachment; filename="foo.txt"',
                'Last-Modified': datetime.now(),
                'X-Accel-Buffering': 'no',
                'Cache-Control': 'no-cache'
            },
        )



TYPE_CHOICES = {
    'AutoField': 'String',
    'BigAutoField': 'Integer',
    'BinaryField': 'String',
    'BooleanField': 'Boolean',
    'CharField': 'String',
    'DateField': 'Dateonly',
    'DateTimeField': 'Date',
    'DecimalField': 'Float',
    'DurationField': 'Number',
    'FileField': 'String',
    'FilePathField': 'String',
    'FloatField': 'Float',
    'IntegerField': 'Integer',
    'BigIntegerField': 'Integer',
    'IPAddressField': 'String',
    'GenericIPAddressField': 'String',
    'JSONField': 'Json',
    'NullBooleanField': 'Boolean',
    'PositiveBigIntegerField': 'Integer',
    'PositiveIntegerField': 'Integer',
    'PositiveSmallIntegerField': 'Integer',
    'SlugField': 'String',
    'SmallAutoField': 'String',
    'SmallIntegerField': 'Integer',
    'TextField': 'String',
    'TimeField': 'Time',
    'UUIDField': 'String',
    'CICharField': 'String',
    'CIEmailField': 'String',
    'CITextField': 'String',
    'HStoreField': 'Json',
}

def handle_one_to_one_field(field):
    if field.target_field.get_internal_type() == 'OneToOneField':
        return TYPE_CHOICES.get(field.target_field.target_field.get_internal_type())
    return TYPE_CHOICES.get(field.target_field.get_internal_type())

def get_type(field):
    # See connection.data_types (different for each DB Engine)
    # ForestAdmin does not handle range fields: https://www.postgresql.org/docs/9.3/rangetypes.html
    # 'RangeField'
    # 'IntegerRangeField'
    # 'BigIntegerRangeField'
    # 'DecimalRangeField'
    # 'DateTimeRangeField'
    # 'DateRangeField'
    if hasattr(field, 'choices') and field.choices is not None:
        return 'Enum'

    field_type = None
    try:
        field_type = field.get_internal_type()
    except AttributeError:
        # Some field as GenericForeignKey hasn't one internal type
        field_type = field.__class__.__name__
    finally:
        # Special case for one to one field which can redirect to an Integer or String
        if field_type == 'OneToOneField':
            return handle_one_to_one_field(field)
        elif field_type == 'ArrayField':
            return [TYPE_CHOICES.get(field.base_field.get_internal_type(), 'unknown')]
        return TYPE_CHOICES.get(field_type, 'unknown')



import copy

from django_forest.utils.schema.definitions import COLLECTION, ACTION, ACTION_FIELD, FIELD
from django_forest.utils.schema import Schema


class Collection:
    name = None  # TODO warn if set to empty string?
    is_searchable = False
    is_read_only = False
    icon = None
    only_for_relationships = None
    pagination_type = None
    search_fields = None
    actions = []
    fields = []
    segments = []

    _registry = {}

    @classmethod
    def register(cls, model_forest, model=None):
        instance = model_forest(model)
        key = instance.__class__.__name__
        if model is not None:
            key = model._meta.db_table
        elif instance.name is not None:
            key = instance.name
        cls._registry[key] = instance

    def load(self):
        pass

    def override_collection(self, collection):
        for key in [k for k in COLLECTION.keys() if k not in ('fields', 'actions', 'segments')]:
            if hasattr(self, key) and getattr(self, key) is not None:
                collection[key] = getattr(self, key)

    def add_smart_field(self, collection, field):
        field['is_virtual'] = True
        field['is_filterable'] = False if 'is_filterable' not in field else field['is_filterable']
        field['is_sortable'] = False if 'is_sortable' not in field else field['is_sortable']
        collection['fields'].append(Schema.get_default(field, FIELD))

    def handle_smart_fields(self, collection):
        existing_fields = [f['field'] for f in collection['fields']]
        for field in self.fields:
            # update
            if field['field'] in existing_fields:
                fi = [f for f in collection['fields'] if f['field'] == field['field']][0]
                fi.update({'is_virtual': True})
            # add
            else:
                self.add_smart_field(collection, field)

    def handle_action_endpoint(self, action):
        if 'endpoint' not in action:
            return f"/forest/actions/{'-'.join(action['name'].split(' ')).lower()}"
        return action['endpoint']

    def handle_action_fields(self, action):
        res = []
        if 'fields' in action:
            for i, v in enumerate(action['fields']):
                field = Schema.get_default(v, ACTION_FIELD)
                field['position'] = i
                res.append(field)
        return res

    def handle_action_hooks(self, action):
        res = {
            'load': False,
            'change': []
        }
        if 'hooks' in action:
            if 'load' in action['hooks']:
                res['load'] = True  # TODO add validation, has to be a function
            if 'change' in action['hooks']:
                res['change'] = list(action['hooks']['change'].keys())
        return res

    def handle_smart_actions(self, collection):
        # TODO, warn if no name?
        for action in self.actions:
            action.update({
                'endpoint': self.handle_action_endpoint(action),
                'fields': self.handle_action_fields(action),
            })
            action_hooks = copy.deepcopy(action)
            action_hooks['hooks'] = self.handle_action_hooks(action_hooks)
            collection['actions'].append(Schema.get_default(action_hooks, ACTION))

    def handle_smart_segments(self, collection):
        for segment in self.segments:
            collection['segments'].append({'name': segment['name']})

    def __init__(self, model):
        self.load()

        # find resource in Schema
        if model is not None:
            collection = Schema.get_collection(model._meta.db_table)
        # create smart collection
        else:
            collection = Schema.get_default({
                'name': self.__class__.__name__,
                'is_virtual': True
            }, COLLECTION)
            Schema.schema['collections'].append(collection)

        if collection is not None:
            self.override_collection(collection)
            self.handle_smart_fields(collection)
            self.handle_smart_actions(collection)
            self.handle_smart_segments(collection)

        super().__init__()
